/*
 * File   : cdproto.c
 * Purpose: This file contains some routines to get or put CDDB entries.
 *          The functions support direct connection (CDDBP) as well as
 *          connections through HTTP-Proxies. With this functionality
 *          direct connection over HTTP are also possible.
 * Author : Andreas Theofilu
 *
 * (C) Copyright 2000 by Andreas Theofilu (Theos Soft)
 */
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

/*
 * The following is used to display a window.
 */
#include <Xmt/Xmt.h>
#include <Xmt/Create.h>
#include <Xmt/Converters.h>
#include <Xmt/Xpm.h>
#include <Xmt/Procedures.h>
#include <Xmt/Layout.h>
#include <Xmt/Chooser.h>
#include <Xmt/InputField.h>
#include <Xmt/WidgetType.h>
#include <Xmt/Dialog.h>
#include <Xmt/Dialogs.h>

#include <Xbae/Matrix.h>
#include <cdaudio.h>

/*
 * This are the program specific headers.
 */
#include "platten.h"
#include "cdrom.h"
#include "config.h"

unsigned long discid;
long total_tracks;
char frame_offsets[512];
long total_seconds;
int nport;
char proxy_name[256];
char cgi_path[256];
int proxy_port;
int PROXY;
Widget cdListWidget = 0;
Widget wquTable = 0;
Widget cdFoundWidget = 0;
Widget wfoTable = 0;
Widget wServer = 0;
Widget wTitle = 0;

struct CUT_S
{
	int pos;
	char title[256];
	char ext[256];
	struct CUT_S *next;
};

struct HEAD_S
{
	char title[256];
	char category[81];
	unsigned long discid;
	struct CUT_S *first;
};

struct HEADLIST_S
{
	char title[256];
	char category[81];
	unsigned long discid;
	struct HEADLIST_S *next;
};

struct HEAD_S head_s;
struct HEADLIST_S *hfirst, *hakt;

typedef struct
{
	char server[81];
	char title[81];
} cdfound_t;

XtResource cdfound_resource[] = {
	{"server", "server", XmtRBuffer, XmtSizeOf (cdfound_t, server),
	 XtOffsetOf (cdfound_t, server), XtRImmediate, NULL},
	{"title", "title", XmtRBuffer, XmtSizeOf (cdfound_t, title),
	 XtOffsetOf (cdfound_t, title), XtRImmediate, NULL}
};

int GetCDDB (Widget w, char *ip);
int ShowQuery (Widget w, char *server);
int FillHead (Widget w, char *str);
int DownloadCD (Widget w, int sockfd, int pos, char *hello, char *ip);
struct CUT_S *GetCutSpace ();
int ListCDs (Widget w, int sockfd);
void FreeList ();
char *CutLF (char *str);
int readn (int fd, char *ptr, int nbytes);
int readline (int fd, char *ptr, int maxlen);
int writen (int fd, char *ptr, int nbytes);
int Connect (char *ip);

/*
 * Folgende Funktion ist die Haupteinstiegsroutine zum Abfragen
 * eines Remoteservers nach der eingelegten CD.
 * Der Funktion wird der Zeiger auf eine Struktur uebergeben,
 * Welche den Titel und die Daten der Tracks festhaelt.
 */
int GetCDDBRemote (Widget w, cddb_title_t *cddb)
{
int ret, i, j, a;
char hv0[128];
struct cddb_conf conf;
struct cddb_serverlist list;
struct cddb_server proxy;
struct disc_info disc;

	XmtDisplayBusyCursor (w);
	memset (cddb, 0, sizeof (cddb_title_t));

	/*
	 * Als erstes lesen wir die Liste der Remoteserver ein.
	 * Diese Liste wird dann Eintrag fuer Eintrag abgearbeitet.
	 */
	if (cddb_read_serverlist (&conf, &list, &proxy) != 0)
        {
           XmtDisplayError (toplevel, "ErrorNoConfig",
		"The configuration file for the remote servers could not\n"
		"be found, or there are no remote servers configured!");
           XmtDisplayDefaultCursor (w);
           return 1;
        }

	/*
	 * Will der User das wir die Daten seiner CD vom
	 * Remoteserver holen?
	 */
	if (conf.conf_access != CDDB_ACCESS_REMOTE)
	{
	   XmtDisplayError (w, "ErrorNoRemote",
		"You've disabled remote CDDB!\n"
		"Enable this option first in the option menu!");
	   XmtDisplayDefaultCursor (toplevel);
	   return 1;
	}

	/*
	 * Wir brauchen die Disk-ID um eine Abfrage starten zu koennen.
	 * Um an diese ID heranzukommen, bemuehen wir eine Funktion der
	 * Bibliothek libcdaudio.
	 * >> Diese Bibliothek sollte ersetzt werden durch eigene <<
	 * >> Funktionen!                                         <<
	 */
	discid = cddbGetDiscID (&disc, app_resources.cdrom, NULL);

	if (discid < 100L || disc.disc_present == 0)
	{
	   XmtDisplayError (w, "ErrorNoCD", NULL);
	   XmtDisplayDefaultCursor (w);
	   return 1;
	}

	total_tracks = disc.disc_total_tracks;
	cddb->tracks = total_tracks;
	cddb->discid = discid;
	memset (&head_s, 0, sizeof (struct HEAD_S));
	hfirst = hakt = NULL;

	/*
	 * Folgende Schleife geht alle Remoteserver durch, bis
	 * entweder ein passender Eintrag gefunden wurde, oder
	 * die Liste zu Ende ist.
	 */
	for (i = 0; i < list.list_len; i++)
	{
	   if (conf.conf_proxy == CDDB_PROXY_ENABLED)
	   {
	      PROXY = 1;
	      strncpy (proxy_name, proxy.server_name, sizeof (proxy_name));
	      proxy_name[sizeof(proxy_name)-1] = 0;
	      proxy_port = proxy.server_port;
	      strncpy (cgi_path, list.list_host[i].host_addressing, sizeof (cgi_path));
	      cgi_path[sizeof(cgi_path)-1] = 0;
	   }
	   else
	      PROXY = 0;

	   if (list.list_host[i].host_protocol == CDDB_MODE_HTTP && !PROXY)
	   {
	      PROXY = 1;
	      strncpy (proxy_name, list.list_host[i].host_server.server_name, sizeof (proxy_name));
	      proxy_name[sizeof(proxy_name)-1] = 0;
	      strncpy (cgi_path, list.list_host[i].host_addressing, sizeof (cgi_path));
	      cgi_path[sizeof(cgi_path)-1] = 0;
	      proxy_port = list.list_host[i].host_server.server_port;
	   }

	   nport = list.list_host[i].host_server.server_port;

	   if (PROXY && list.list_host[i].host_protocol == CDDB_MODE_CDDBP)
	      continue;

	   total_seconds = disc.disc_length.minutes * 60 + disc.disc_length.seconds;
	   /*
	    * Zusammensetzen der Frame offsets. Das muss jedesmal neu
	    * gemacht werden um die Zeile an das Protokoll anzupassen.
	    */
	   frame_offsets[0] = 0;

	   for (j = 0; j < total_tracks; j++)
	   {
	      if (j > 0)
		 strcat (frame_offsets, (PROXY) ? "+" : " ");

	      sprintf (hv0, "%d", disc.disc_track[j].track_lba);
	      strcat (frame_offsets, hv0);
	   }

	   ret = GetCDDB (w, list.list_host[i].host_server.server_name);

	   if (!ret)
	   {
	   struct CUT_S *akt;

	      if (!ShowQuery (w, list.list_host[i].host_server.server_name))
	      {
                 XmtDisplayDefaultCursor (w);
	         return 1;
	      }

	      strncpy (cddb->title, head_s.title, 80);
	      a = 0;
	      akt = head_s.first;

	      while (akt && a < 30)
	      {
	         strncpy (cddb->entry[a].song, akt->title, 80);
		 strncpy (cddb->entry[a].singer, akt->ext, 80);
		 akt = akt->next;
		 a++;
	      }

	      FreeList ();
              XmtDisplayDefaultCursor (w);
	      return 0;
	   }

	   FreeList ();

	   if (ret > 0)
	   {
              XmtDisplayDefaultCursor (w);
	      return 1;
	   }
	}

        XmtDisplayDefaultCursor (w);
	return 0;
}

/*
 * Folgende Funktion erzeugt ein Fenster in dem das Ergebnis des
 * Downloads der CD zu sehen ist. Der User kann entscheiden ob er
 * Die Daten in seiner Datenbank haben will oder nicht.
 *
 * Eventuell sollte der User hier die Gelegenheit haben die Daten
 * gleich zu editieren?? Macht das wer?
 */
int ShowQuery (Widget w, char *server)
{
char hv0[81], hv1[81];
String *row, s;
int i;
struct CUT_S *akt;
cdfound_t cdfound;

	if (!cdFoundWidget)
	{
	   if (!GetResourcePath ("CDFound"))
	   {
	      XmtDisplayError (w, "ErrorNoWidgets", NULL);
	      return 0;
	   }

	   cdFoundWidget = XmtBuildQueryDialog (w, "cdfound",
		cdfound_resource, XtNumber (cdfound_resource),
		"server", &wServer, "title", &wTitle,
		"foTable", &wfoTable, NULL);
	   XmtAddDeleteCallback (XmtGetShell (cdFoundWidget), XmDO_NOTHING,
		XmtDialogCancelCallback, NULL);
	}

	/*
	 * Hier fuellen wir die ersten beiden Zeilen der
	 * Dialogbox.
	 */
	memset (&cdfound, 0, sizeof (cdfound_t));
	strncpy (cdfound.server, server, 80);
	strncpy (cdfound.title, CutLF (head_s.title), 80);

	/*
	 * Nun fuellen wir noch die Tabelle in der Dialogbox.
	 */
	XbaeMatrixDisableRedisplay (wfoTable);
	XbaeMatrixDeleteRows (wfoTable, 0, XbaeMatrixNumRows (wfoTable));
	akt = head_s.first;
	i = 0;

	while (akt)
	{
	   row = (String *)XtMalloc (2 * sizeof (String));
	   memset (&hv0[0], 0, sizeof (hv0));
	   memset (&hv1[0], 0, sizeof (hv1));
	   strncpy (&hv0[0], CutLF (akt->title), 80);
	   strncpy (&hv1[0], CutLF (akt->ext), 80);
	   row[0] = &hv0[0];
	   row[1] = &hv1[0];
	   XbaeMatrixAddRows (wfoTable, i, row, NULL, NULL, 1);
	   XtFree ((char *)row);
	   akt = akt->next;
	   i++;
	}

	XbaeMatrixEnableRedisplay (wfoTable, True);

	if (!XmtDialogDoSync (cdFoundWidget, &cdfound))
	   return 0;

	/*
	 * Da der User Daten geaendert haben koennte, uebernehmen
	 * wir die in Frage kommenden Felder aus der Dialogbox.
	 */
	strcpy (head_s.title, cdfound.title);
	i = 0;
	akt = head_s.first;

	while (akt)
	{
	   s = XbaeMatrixGetCell (wfoTable, i, 0);
	   strcpy (akt->title, s);
	   s = XbaeMatrixGetCell (wfoTable, i, 1);
	   strcpy (akt->ext, s);
	   akt = akt->next;
	   i++;
	}

	return 1;
}

/*
 * Folgende Funktion wird aufgerufen, sobald der User in eine der Zellen
 * der Tabelle in der Dialogbox klickt. Die Funktion hebt zunaechst die
 * Markierung der einzelnen Zelle auf um dann die ganze Zeile zu markieren.
 */
void SelectCDFoundCallback (Widget w, XtPointer tag, XtPointer data)
{
XbaeMatrixEnterCellCallbackStruct *ec = (XbaeMatrixEnterCellCallbackStruct *)data;
	ec->doit = False;
	ec->map = False;
	XbaeMatrixDeselectAll (w);
	XbaeMatrixSelectRow (w, ec->row);
}

/*
 * Folgende Funktion wird aufgerufen nach dem der User eine eben
 * editierte Zelle verlassen hat.
 */
void cdFoundLeaveCellCallback (Widget w, XtPointer tag, XtPointer data)
{
XbaeMatrixLeaveCellCallbackStruct *lc = (XbaeMatrixLeaveCellCallbackStruct *)data;

	lc->doit = True;

	if (lc->column == 1)
	{
	   if (!strlen (lc->value))
	   {
	      XmtDisplayError (w, "ErrorEmTitle", NULL);
	      lc->doit = False;
	      return;
	   }

	   *lc->value = toupper (*lc->value);
	}

	if (isalpha (*lc->value))
	   *lc->value = toupper (*lc->value);
}

/*
 * Folgende Funktion wird durch einen Button im Dialogfenster
 * aufgerufen und fuellt den Inhalt der ersten Zelle in alle
 * weiteren Zellen der selben Spalte, welche noch leer sind.
 */
void cdFoundFillCallback ()
{
String s, p;
int i, j;

	s = XbaeMatrixGetCell (wfoTable, 0, 1);

	if (!strlen (s))
	   return;

	i = XbaeMatrixNumRows (wfoTable);
	XbaeMatrixDisableRedisplay (wfoTable);

	for (j = 1; j < i; j++)
	{
	   p = XbaeMatrixGetCell (wfoTable, j, 1);

	   if (!strlen (p))
	      XbaeMatrixSetCell (wfoTable, j, 1, s);
	}

	XbaeMatrixEnableRedisplay (wfoTable, True);
}

/*
 * Folgende Funktion stellt eine Verbindung mit einem Remoteserver
 * her, dessen Name per Parameter uebergeben wird. Wird ein Eintrag
 * gefunden, werden die entsprechenden Strukturen aufgefuellt.
 */
int GetCDDB (Widget w, char *ip)
{
int sockfd;
char hv0[1024], hv1[256], hv2[256], hv3[256], hv4[256], *p;
char hello[512];
int i, code;
struct hostent *he;

	if ((sockfd = Connect (ip)) < 0)
	   return 1;

	code = 0;
	p = NULL;

	/*
	 * Als erstes muessen wir mit dem Server eine Verbindung
         * haben. Kommt diese zustande, meldet sich der Server
         * mit einem Banner. Dieses muss ausgewertet werden um zu
         * erkennen ob es wirklich ein CDDB-Server ist.
	 */
	if (!PROXY)
	{
	   i = read (sockfd, &hv0[0], sizeof (hv0));
	   hv0[i] = 0;
	   p = strtok (hv0, " ");
	   i = 0;

	   while (p)
	   {
	      switch (i)
	      {
	         case 0:	/* Return code */
		    code = atoi (p);
	         break;

	         case 1:	/* Hostname */
		    strncpy (hv1, p, sizeof (hv1));
	         break;

	         case 2:	/* Version */
		    strncpy (hv2, p, sizeof (hv2));
	         break;

	         /* Rest is the current date and time. */
	      }

	      p = strtok (NULL, " ");
	      i++;
	   }

/****	   switch (code)
	   {
	      case 200: printf ("OK, read/write allowed\n"); break;
	      case 201: printf ("OK, read only\n"); break;
	      case 432: printf ("No connections allowed: permission denied\n"); break;
	      case 433: printf ("No connections allowed: X users allowed, Y currently active\n");
	      case 434: printf ("No connections allowed: system load too high\n");
	   }
****/
	   if (code < 200 || code > 201)
	   {
	      close (sockfd);
	      sprintf (hv0, "The following error occured:\nError %d: %s",
		code, (p != NULL) ? p : "Unknown error");
	      XmtDisplayError (w, NULL, hv0);
	      return 1;
	   }
	}

	/*
	 * Der Server hat sich also gemeldet und erlaubt uns zumindest
	 * zu lesen. Bevor wir das machen koennen, muessen wir erst
	 * einmal einen Handshake machen.
	 * Dazu benoetigen wir einen nach Moeglichkeit vollstaendigen
	 * Domainnamen.
	 */
	if (gethostname (&hv1[0], sizeof (hv1)) < 0 ||
		(he = gethostbyname (&hv1[0])) == NULL)
	{
	   strcpy (hv1, "unknown");
	}
	else
	{
	   if (strchr (he->h_name, '.') != NULL)
	   {
	      strncpy (hv1, he->h_name, sizeof (hv1));
	   }
	   else
	   {
	      char **ap;

	      for (ap = he->h_aliases; *ap != NULL; ap++)
	      {
		 if (strchr (*ap, '.') != NULL)
		 {
		    strncpy (hv1, *ap, sizeof (hv1));
		    break;
		 }
	      }
	   }
	}

	hv1[sizeof(hv1)-1] = 0;
	/*
	 * Folgender String wird nur benoetigt wenn wir HTTP-Protokoll
	 * verwenden. Verwenden wir einen echten Proxy, kann NUR HTTP-
	 * Protokoll angewendet werden.
	 */
	sprintf (hello, "hello=user+%s+RecordManager+%s&proto=2",
		hv1, VERSION);

	if (!PROXY)
	{
	   sprintf (hv0, "cddb hello user %s RecordManager %s\n", hv1, VERSION);
	   writen (sockfd, hv0, strlen (hv0));
	   memset (hv0, 0, sizeof (hv0));
	   i = read (sockfd, &hv0[0], sizeof (hv0));
	   hv0[i] = 0;
	   p = strtok (hv0, " ");
	   i = code = 0;

	   while (p)
	   {
	      switch (i)
	      {
	         case 0:	/* Return code */
		    code = atoi (p);
	         break;

	         case 1:	/* hello and welcome */
	         case 2:
                 case 3:
		    strncpy (hv1, p, sizeof (hv1));
	         break;

	         case 4:	/* username@hostname */
		    strncpy (hv2, p, sizeof (hv2));
	         break;

	         case 5:	/* running clientname */
	         case 6:
		    strncpy (hv3, p, sizeof (hv3));
	         break;

	         case 7:	/* version */
	            strncpy (hv4, p, sizeof (hv4));
	         break;
	      }

	      if (code < 200 || code > 999)
	         break;

	      p = strtok (NULL, " ");
	      i++;
	   }

/**	   switch (code)
	   {
	      case 200: printf ("Handshake successful\n"); break;
	      case 431: printf ("Handshake not successful, closing connection\n"); break;
	      case 402: printf ("Already shook hands\n"); break;
	   }
***/
	   if (code != 200 && code != 402)
	   {
	      close (sockfd);
	      sprintf (hv0, "The following error occured:\nError %d: %s",
		code, (p != NULL) ? p : "Unknown Error");
	      XmtDisplayError (w, NULL, hv0);
	      return 1;
	   }
	}

	/*
	 * Nun endlich koennen wir eine Suche absetzen.
	 */
	if (PROXY)
	   sprintf (hv0, "cmd=CDDB+QUERY+%08lx+%ld+%s+%ld", discid,
		total_tracks, frame_offsets, total_seconds);
	else
	   sprintf (hv0, "cddb query %08lx %ld %s %ld\n", discid, total_tracks,
		frame_offsets, total_seconds);

	if (!PROXY)
	{
	   writen (sockfd, hv0, strlen (hv0));
	   i = readline (sockfd, hv0, sizeof (hv0));
	   hv0[i] = 0;

	   if ((p = strchr (hv0, ' ')) == NULL)
	   {
	      close (sockfd);
	      XmtDisplayError (w, "ErrorResponse",
		"Got invalid response from server!");
	      return 1;
	   }

	   *p = 0;
	   code = atoi (hv0);
	}
	else
	{
	   sprintf (hv1, "GET http://%s:%d%s?%s&%s HTTP/1.0 \n\n",
		ip, nport, cgi_path, hv0, hello);
	   writen (sockfd, hv1, strlen (hv1));
	   code = 0;

	   while ((i = readline (sockfd, &hv0[0], sizeof (hv0))) > 0)
	   {
	      if ((p = strchr (hv0, ' ')) != NULL)
	      {
	         *p = 0;
	         code = atoi (hv0);
	      }
	   }

	   /*
	    * Der Proxy beendet die Kommunikation hier ohnehin,
	    * also schliessen wir den socket auch.
	    */
	   if (code != 211)
	      close (sockfd);
	}

/****	switch (code)
	{
	   case 200: printf ("Found exact match\n"); break;
	   case 211: printf ("Found inexact matches, list follows...\n"); break;
	   case 202: printf ("No match found\n"); break;
	   case 403: printf ("Database entry is corrupt\n"); break;
	   case 409: printf ("No handshake\n"); break;
	   case 500: printf ("General error: %s\n", p+1); break;
	}
****/

	if (code > 299)
	{
	   sprintf (hv0, "The following error occured:\nError %d: %s",
		code, (p != NULL) ? p+1 : "Unknown Error");
	   XmtDisplayError (w, NULL, hv0);
	   return 1;
	}

	if (code == 200)	/* Exact match? */
	{
	   p++;

	   if (FillHead (w, p))
	   {
	      close (sockfd);
	      return 1;
	   }

	   if (PROXY)
	   {
	      if ((sockfd = Connect (ip)) < 0)
		 return 1;
	   }

	   if (DownloadCD (w, sockfd, 0, hello, ip))
	   {
	      close (sockfd);
	      return 1;
	   }
	}
	else if (code == 211)
	{
	   if (ListCDs (w, sockfd))
	   {
	      if (!PROXY)
	         close (sockfd);

	      return 1;
	   }

	   if (PROXY)
	   {
	      if ((sockfd = Connect (ip)) < 0)
	         return 1;
	   }

	   if (DownloadCD (w, sockfd, i, hello, ip))
	   {
	      close (sockfd);
	      return 1;
	   }
	}

	close (sockfd);

	if (code == 202)	/* No match found */
	   return -1;

	return 0;
}

/*
 * Folgende Funktion liest den CD-Header ein und fuegt die
 * Daten in eine Struktur ein. Die Funktion wird aufgerufen,
 * sobald ein gueltiger CD Eintrag gefunden wurde.
 */
int FillHead (Widget w, char *str)
{
char *p, *s;

	s = str;

	if ((p = strchr (s, ' ')) == NULL)
	{
	   XmtDisplayError (w, "ErrorCategory",
		"No category found!");
	   return 1;
	}

	*p = 0;
	strncpy (head_s.category, s, 80);
	head_s.category[80] = 0;
	p++;
	s = p;

	if ((p = strchr (s, ' ')) == NULL)
	{
	   XmtDisplayError (w, "ErrorDiscID",
		"No discID found!");
	   return 1;
	}

	*p = 0;
	head_s.discid = strtoul (s, NULL, 16);

	p++;
	s = p;
	strncpy (head_s.title, s, 255);
	head_s.title[255] = 0;
	CutLF (head_s.title);

	if (head_s.first)
	{
	   struct CUT_S *first, *akt;

	   first = akt = head_s.first;

	   while (akt)
	   {
	      akt = first->next;
	      XtFree ((char *)first);
	      first = akt;
	   }

	   head_s.first = NULL;
	}

	return 0;
}

/*
 * Folgende Funktion laedt die Daten einer CD vom Remoteserver
 * herunter und fuegt die Daten in eine interne Struktur ein.
 * Diese Funktion ist die letzte Konsequenz der gesamten Suche
 * nach einer CD und bedeutet in jedem Fall das eine CD
 * gefunden wurde.
 */
int DownloadCD (Widget w, int sockfd, int pos, char *hello, char *ip)
{
char hv0[1024], hv1[256], *p;
int i, j, code;
struct CUT_S *akt;

	p = NULL;
	akt = NULL;

	if (PROXY)
	   sprintf (hv0, "GET http://%s:%d%s?cmd=CDDB+READ+%s+%08lx&%s HTTP/1.0 \n\n",
		ip, nport, cgi_path, head_s.category, head_s.discid, hello);
	else
	   sprintf (hv0, "cddb read %s %08lx\n", head_s.category, head_s.discid);

	writen (sockfd, hv0, strlen (hv0));

	if (!PROXY)
	{
	   do
	   {
	      i = readline (sockfd, &hv0[0], sizeof (hv0));
	   }
	   while ((hv0[0] == '\n' || hv0[0] == '\r') && i != 0);

	   if ((p = strchr (hv0, ' ')) == NULL)
	   {
	      XmtDisplayError (w, "ErrorSyntax",
		"Command syntax error!");
	      return 1;
	   }

	   *p = 0;
	   p++;
	   code = atoi (hv0);
	}
	else
	{
	   code = 0;

	   while ((i = readline (sockfd, &hv0[0], sizeof (hv0))) > 0)
	   {
	      if ((p = strchr (hv0, ' ')) != NULL)
	      {
		 *p = 0;
		 code = atoi (hv0);
	      }

	      if (code >= 200 && code < 600)
	         break;
	   }
	}

/***	switch (code)
	{
	   case 210: printf ("OK, CDDB database entry follows\n"); break;
	   case 401: printf ("Specified CDDB entry not found.\n"); break;
	   case 402: printf ("Server error.\n"); break;
	   case 403: printf ("Database entry is corrupt.\n"); break;
	   case 409: printf ("No handshake.\n"); break;
	}
***/
	if (code != 210)
	{
	   sprintf (hv0, "The following error occured:\nError %d: %s",
		code, (p != NULL) ? p+1 : "Unknown error");
	   XmtDisplayError (w, NULL, hv0);
	   return 1;
	}

	j = 0;

	while ((i = readline (sockfd, &hv0[0], sizeof (hv0))) > 0)
	{
	   hv0[i] = 0;
	   CutLF (&hv0[0]);

	   if (strlen (hv0) == 1 && hv0[0] == '.')	/* EOF? */
	      break;

	   if (hv0[0] == '#')				/* Remark */
	      continue;

	   if (!strlen (hv0) || hv0[0] == '\n' || hv0[0] == '\r')
	      continue;

	   if ((p = strchr (hv0, '=')) != NULL)
	   {
	      *p = 0;
	      p++;

	      sprintf (hv1, "TTITLE%d", j);

	      if (strcmp (hv0, hv1) == 0)
	      {
		 if ((akt = GetCutSpace ()) == NULL)
		 {
		    FreeList ();
		    return 1;
		 }

		 strncpy (akt->title, p, 255);
		 akt->pos = j;
		 j++;
		 continue;
	      }

	      if (strcmp (hv0, "EXTD") == 0)
	      {
	         j = 0;
	         akt = head_s.first;

		 if (akt == NULL)
		 {
		    XmtDisplayError (w, NULL, "Fatal internal error!\n"
			"Please inform Andreas Theofilu per Mail "
			"(theofilu@@aon.at). TIA.");
		    FreeList ();
		    return 1;
		 }

	         continue;
	      }

	      sprintf (hv1, "EXTT%d", j);

	      if (strcmp (hv0, hv1) == 0 && akt->pos == j)
	      {
		 strncpy (akt->ext, p, 255);
	         akt = akt->next;
		 j++;
	      }
	   }
	}

	return 0;
}

struct CUT_S *GetCutSpace ()
{
struct CUT_S *akt, *last;

	if (head_s.first == NULL)
	{
	   if ((akt = (struct CUT_S *)XtMalloc (sizeof (struct CUT_S))) == NULL)
	   {
	      XmtDisplayError (toplevel, "ErrorNoMem", NULL);
	      return NULL;
	   }

	   memset (akt, 0, sizeof (struct CUT_S));
	   head_s.first = akt;
	}
	else
	{
	   last = akt = head_s.first;

	   while (akt)
	   {
	      akt = akt->next;

	      if (akt)
		 last = akt;
	   }

	   if ((akt = (struct CUT_S *)XtMalloc (sizeof (struct CUT_S))) == NULL)
	   {
	      XmtDisplayError (toplevel, "ErrorNoMem", NULL);
	      return NULL;
	   }

	   memset (akt, 0, sizeof (struct CUT_S));
	   last->next = akt;
	}

	return akt;
}

/*
 * Folgende Funktion wird immer dann aufgerufen wenn die Suche
 * mehr als eine CD gefunden hat. In diesem Fall wird dem User
 * ein Dialogfenster praesentiert, aus dem er die gewuenschte
 * CD auswaehlen kann.
 */
int ListCDs (Widget w, int sockfd)
{
char *p, *s, hv0[1024];
int i;

	while ((i = readline (sockfd, &hv0[0], sizeof (hv0))) > 0)
	{
	   /*
	    * Ende erreicht?
	    */
	   if (i == 1 && hv0[0] == '.')
	      break;

	   if (!hfirst)
	   {
	      if ((hfirst = (struct HEADLIST_S *)XtMalloc (sizeof (struct HEADLIST_S))) == NULL)
	      {
	         fprintf (stderr, "Error!\nNot enough memory to hold the header information!\n");
	         return 1;
	      }

	      memset (hfirst, 0, sizeof (struct HEADLIST_S));
	      hakt = hfirst;
	   }
	   else
	   {
	      struct HEADLIST_S *neu;

	      if ((neu = (struct HEADLIST_S *)XtMalloc (sizeof (struct HEADLIST_S))) == NULL)
	      {
	         fprintf (stderr, "Error!\nNot enough memory to hold the header information!\n");
	         return 1;
	      }

	      memset (neu, 0, sizeof (struct HEADLIST_S));
	      hakt->next = neu;
	      hakt = neu;
	   }

	   /*
	    * Zerlegen der Zeile und pruefen.
	    */
	   s = &hv0[0];

	   if ((p = strchr (s, ' ')) == NULL)
	   {
	      fprintf (stderr, "Error!\nSyntax Error in response!\n");
	      FreeList ();
	      return 1;
	   }

	   *p = 0;
	   strncpy (hakt->category, s, 80);
	   s = p + 1;

	   if ((p = strchr (s, ' ')) == NULL)
	   {
	      fprintf (stderr, "Error!\nSyntax Error in response!\n");
	      FreeList ();
	      return 1;
	   }

	   *p = 0;
	   hakt->discid = strtoul (s, NULL, 16);
	   s = p + 1;
	   strncpy (hakt->title, s, 255);

//	   fprintf (stderr, "%08lx | %-20s | %s\n",
//		hakt->discid, hakt->category, hakt->title);
	}

	/*
	 * An dieser Stelle zeigen wir unser mehrdeutiges Suchergebnis
	 * in einem Fenster an, um den Anwender auswaehlen zu lassen.
	 */
	XmtDisplayBusyCursor (w);

        if (!cdListWidget)
        {
           GetResourcePath ("CDList");
           cdListWidget = XmtBuildQueryDialog (w, "cdlist",
                NULL, 0, "quTable", &wquTable, NULL);

           XmtAddDeleteCallback (XmtGetShell (cdListWidget), XmDO_NOTHING,
                XmtDialogCancelCallback, NULL);
        }
        else
        {
           Widget list = XmtNameToWidget (toplevel, "*cdlist*quTable");
           FirstCDListFillCallback (cdListWidget, list);
        }

        i = XmtDialogDoSync (cdListWidget, NULL);
	XmtDisplayDefaultCursor (w);

	if (!i)
	   return 1;

	return 0;
}

/*
 * Folgende Funktion wird aufgerufen, sobald der User in eine der Zellen
 * der Tabelle in der Dialogbox klickt. Die Funktion hebt zunaechst die
 * Markierung der einzelnen Zelle auf um dann die ganze Zeile zu markieren.
 */
void SelectCDListCallback (Widget w, XtPointer tag, XtPointer data)
{
int i = 0;
XbaeMatrixEnterCellCallbackStruct *ec = (XbaeMatrixEnterCellCallbackStruct *)data;

	ec->doit = False;
	ec->map = False;
	XbaeMatrixDeselectAll (w);
	XbaeMatrixSelectRow (w, ec->row);
	hakt = hfirst;

	while (i < ec->row)
	{
	   hakt = hakt->next;
	   i++;
	}

	strcpy (head_s.title, hakt->title);
	strcpy (head_s.category, hakt->category);
	head_s.discid = hakt->discid;
}

/*
 * Folgende Funktion wird aufgerufen nach dem die Dialogbox zum Anzeigen
 * der gefundenen CDs erzeugt wurde, aber bevor die Dialogbox sichtbar
 * ist. Zu diesem Zeitpunkt sind alle Widgets bereits erzeugt und
 * initialisiert.
 */
void FirstCDListFillCallback (Widget dialog, Widget list)
{
char hv0[128];
String *row;
int i = 0;

	hakt = hfirst;
	XbaeMatrixDisableRedisplay (list);
	XbaeMatrixDeleteRows (list, 0, XbaeMatrixNumRows (list));

	while (hakt)
	{
	   row = (String *)XtMalloc (3 * sizeof (String));
	   sprintf (hv0, "%08lx", hakt->discid);
	   row[0] = &hv0[0];
	   row[1] = hakt->category;
	   row[2] = hakt->title;
	   XbaeMatrixAddRows (list, i, row, NULL, NULL, 1);
	   XtFree ((char *)row);
	   hakt = hakt->next;
	   i++;
	}

	XbaeMatrixEnableRedisplay (list, True);
	XbaeMatrixSelectRow (list, 0);
	hakt = hfirst;
	strcpy (head_s.title, hakt->title);
	strcpy (head_s.category, hakt->category);
	head_s.discid = hakt->discid;
}

void FreeList ()
{
struct CUT_S *akt, *first;

	hakt = hfirst;

	while (hakt)
	{
	   hakt = hfirst->next;
	   XtFree ((char *)hfirst);
	   hfirst = hakt;
	}

	hakt = hfirst = NULL;

	akt = first = head_s.first;

	while (akt)
	{
	   akt = first->next;
	   XtFree ((char *)first);
	   first = akt;
	}

	head_s.first = NULL;	
}

char *CutLF (char *str)
{
char *p = str;

	while (*p)
	{
	   if (*p == '\n' || *p == '\r')
	   {
	      *p = 0;
	      break;
	   }

	   p++;
	}

	return str;
}

int readn (int fd, char *ptr, int nbytes)
{
int nleft, nread;

	nleft = nbytes;

	while (nleft > 0)
	{
	   nread = read (fd, ptr, nleft);

	   if (nread < 0)
	      return nread;
	   else if (nread == 0)
	      break;

	   nleft -= nread;
	   ptr   += nread;
	}

	return (nbytes - nleft);
}

int readline (int fd, char *ptr, int maxlen)
{
int n, rc;
char c;

	for (n = 1; n < maxlen; n++)
	{
	   if ((rc = read (fd, &c, 1)) == 1)
	   {
	      *ptr++ = c;

	      if (c == '\n')
		 break;

	      if (c == '.' && n == 1)
	         break;
	   }
	   else if (rc == 0)
	   {
	      if (n == 1)
		 return 0;
	      else
		 break;
	   }
	   else
	      return -1;
	}

	*ptr = 0;
	return n;
}

int writen (int fd, char *ptr, int nbytes)
{
int nleft, nwritten;

	nleft = nbytes;

	while (nleft > 0)
	{
	   nwritten = write (fd, ptr, nleft);

	   if (nwritten <= 0)
	      return nwritten;

	   nleft -= nwritten;
	   ptr   += nwritten;
	}

	return (nbytes - nleft);
}

int Connect (char *ip)
{
int sockfd;
struct hostent *hostptr;
struct in_addr *ptr;
struct sockaddr_in serv_addr;

	if (!PROXY)
	{
	   if ((hostptr = gethostbyname (ip)) == NULL)
	   {
	      fprintf (stderr, "Error!\nIllegal name >%s<!\n", ip);
	      return 1;
	   }
	}
	else
	{
	   if ((hostptr = gethostbyname (proxy_name)) == NULL)
	   {
	      fprintf (stderr, "Error!\nIllegal proxy >%s<!\n", proxy_name);
	      return 1;
	   }
	}

	ptr = (struct in_addr *)hostptr->h_addr_list[0];
        memset ((char *)&serv_addr, 0, sizeof (serv_addr));
        serv_addr.sin_family = AF_INET;
        serv_addr.sin_addr.s_addr = ptr->s_addr;

	if (!PROXY)
           serv_addr.sin_port = htons (nport);
	else
	   serv_addr.sin_port = htons (proxy_port);

        if ((sockfd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
        {
           fprintf (stderr, "Error!\nCan't open stream socket!\n");
           return -1;
        }

        if (connect (sockfd, (struct sockaddr *)&serv_addr,
                        sizeof (serv_addr)) < 0)
        {
           fprintf (stderr, "Error!\nCan't connect to server!\n");
           close (sockfd);
	   return -1;
        }

	return sockfd;
}

